package net.sergeych.collections

/**
 * Automatically mutable sorted list based on binary search and a given comparing function.
 * To construct list of comparable elements use [invoke]. There is a secondary constructor
 * to use with existing [Comparator] instance.
 *
 * While the sorted list is mutable, it does not implement `MutableList` because indexed
 * assignments are not possible keeping sort order; use [add] instead.
 *
 * It is possible to store several equal values and retrieve them all. See [add] and [addIfNotExists].
 */
class SortedList<T: Any>(
    private val list: MutableList<T> = mutableListOf(),
    private val compare: (T,T) -> Int
)
    : List<T> by list
{
    @Suppress("unused")
    constructor(list: MutableList<T>, comparator: Comparator<T>)
            : this(list,{ a, b -> comparator.compare(a,b) })

    private fun binarySearch(element: T): Int {
        var low = 0
        var high = this.size - 1

        while (low <= high) {
            val mid = (low + high).ushr(1) // unsigned shift right, equivalent to integer division of sum by 2
            val midVal = list[mid]
            val cmp = compare(element,midVal)

            when {
                cmp < 0 -> high = mid - 1
                cmp > 0 -> low = mid + 1
                else -> return mid // key found
            }
        }
        return -(low + 1) // key not found, insertion point is -(low + 1)
    }

    /**
     * Find any element equals to value using fast binary search.
     *
     * Note that if there are many elements that are equal to [value] using the [compare],
     * it will return index of some of it. Use [findFirst] and [findLast] if needed.
     *
     * @return found value or null
     */
    fun find(value: T): T? {
        val i = binarySearch(value)
        return if( i < 0 ) null else list[i]
    }

    /**
     * Find all elements equal to the value.
     * @return list of found elements, or an empty list.
     */
    fun findAll(value: T): List<T> {
        val result = mutableListOf<T>()
        val start = binarySearch(value)
        if( start >= 0) {
            for( i in start ..< size ) {
                val element = list[i]
                if( compare(value, element) == 0 )
                    result += element
                else
                    break
            }
            if( start > 0) {
                for( i in (start-1) downTo 0) {
                    val element = list[i]
                    if (compare(value, element) == 0)
                        result += element
                    else
                        break
                }
            }
        }
        return result
    }

    /**
     * Add all values. Duplicates will also be added.
     */
    fun add(vararg values: T) {
        for( value in values ) {
            val i = binarySearch(value)
            if (i >= 0)
                list.add(i + 1, value)
            else
                list.add(-(i + 1), value)
        }
    }

    /**
     * Remove one element equals to value.
     * @return true if the element has been removed
     */
    fun remove(value: T): Boolean {
        val i = binarySearch(value)
        return if( i >= 0) {
            list.removeAt(i)
            true
        }
        else false
    }

    /**
     * Remove element at index.
     * @return element that has been removed
     */
    @Suppress("unused")
    fun removeAt(index: Int): T = list.removeAt(index)

    /**
     * Optimized, binary search based version.
     * @returns index of the _first_ occurrence of the element, or -1
     */
    override fun indexOf(element: T): Int {
        var i = binarySearch(element)
        if( i < 0 ) return -1
        while( i > 0 && compare(element, list[i-1]) == 0) i--
        return i
    }

    /**
     * Optimized, binary search based version.
     * @returns index of the _last_ occurrence of the element, or -1
     */
    override fun lastIndexOf(element: T): Int {
        var i = binarySearch(element)
        if( i < 0 ) return  -1
        while( i < list.size && compare(element, list[i+1])==0) i++
        return i
    }

    /**
     * Optimized, binary search based, first occurrence of the element.
     *
     * Note that the order of 'equal' elements is unspecified, order of appearance is not kept.
     */
    fun findFirst(element: T): T? {
        val i = indexOf(element)
        return if( i < 0 ) null else list[i]
    }

    /**
     * Optimized, binary search based search of the last occurrence of the element
     *
     * Note that the order of 'equal' elements is unspecified, order of appearance is not kept.
     */
    fun findLast(element: T): T? {
        val i = lastIndexOf(element)
        return if( i < 0 ) null else list[i]
    }

    /**
     * Remove all elements equals to value.
     * @return number of removed elements
     */
    fun removeAll(value: T): Int {
        var start = binarySearch(value)
        var count = 0
        while( start < size && compare(value, list[start]) == 0 ) {
            list.removeAt(start)
            count++
        }
        while( start > 0 && compare(value, list[--start]) == 0) {
            list.removeAt(start)
            count++
        }
        return count
    }

    override operator fun contains(element: T): Boolean = binarySearch(element) >= 0

    /**
     * Add a value if it is not yet in the list.
     * @return true if the value was added and false if it is already in the list, and was not added.
     */
    fun addIfNotExists(value: T): Boolean {
        val i = binarySearch(value)
        return if( i < 0) {
            list.add(-(i + 1), value)
            true
        }
        else false

    }

    companion object {
        /**
         * Construct list of elements from comparable instances.
         */
        operator fun <T: Comparable<T>>invoke(vararg values: T): SortedList<T> =
            SortedList(values.toList().sorted().toMutableList()) { a, b -> a.compareTo(b) }
    }

}